// gps_latlong.ino
//
// 緯度、経度を測る（GPSモジュール使用）
// 
// Written by Nobuki HIRAMINE
// Last modified : 2011/12/03

#include <LiquidCrystal.h>
#include <SoftwareSerial.h>

#define READBUFFERSIZE	(256)
#define DELIMITER		(",")

LiquidCrystal g_lcd( 12, 11, 5, 4, 3, 2 );
SoftwareSerial g_gps( 6, 7 );
char g_szReadBuffer[READBUFFERSIZE] = "";
int  g_iIndexChar = 0;

void setup()
{
	g_gps.begin(9600);

	g_lcd.begin( 16, 2 );
	g_lcd.clear();
}

// センテンスの解析。
// $GPRMCの場合、引数変数に、緯度、経度を入れ、戻り値 1 を返す。
// $GPRMC以外の場合、戻り値は 0 を返す。
int AnalyzeLineString( char szLineString[],
					   float& rfLat, float& rfLong )
{
	rfLat = 0.0;
	rfLong = 0.0;
	
	// $GPRMC
	if( 0 != strncmp( "$GPRMC", szLineString, 6 ) )
	{
		return 0;
	}
	
	// $GPRMC,085120.307,A,3541.1493,N,13945.3994,E,000.0,240.3,181211,,,A*6A
	strtok( szLineString, DELIMITER );	// $GPRMC
	strtok( NULL, DELIMITER );	// UTC時刻
	strtok( NULL, DELIMITER );	// ステータス
	char* pszLat = strtok( NULL, DELIMITER );	// 緯度(dddmm.mmmm)
	strtok( NULL, DELIMITER );	// 北緯か南緯か
	char* pszLong = strtok( NULL, DELIMITER );	// 経度(dddmm.mmmm)
	
	if( NULL == pszLong )
	{
		return 0;
	}
	
	// dddmm.mmmm → ddd.dddddd
	float temp, deg, min;

	temp = atof(pszLat);
	deg = (int)(temp/100);
	min = temp - deg * 100;
	rfLat = deg + min / 60;
	
	temp = atof(pszLong);
	deg = (int)(temp/100);
	min = temp - deg * 100;
	rfLong = deg + min / 60;
	
	return 1;
}

// １行文字列の読み込み
// 0 : 読み取り途中。1 : 読み取り完了。
int ReadLineString(	SoftwareSerial& serial,
					char szReadBuffer[], const int ciReadBufferSize, int& riIndexChar,
					char szLineString[], const int ciLineStringSize )
{
	while( 1 )
	{
		char c = serial.read();
		if( -1 == c )
		{
			break;
		}
		if( '\r' == c  )
		{	// 終端
			szReadBuffer[riIndexChar] = '\0';
			strncpy( szLineString, szReadBuffer, ciLineStringSize - 1 );
			szLineString[ciLineStringSize - 1] = '\0';
			riIndexChar = 0;
			return 1;
		}
		else if( '\n' == c )
		{
			;	// 何もしない
		}
		else
		{	// 途中
			if( (ciReadBufferSize - 1) > riIndexChar )
			{
				szReadBuffer[riIndexChar] = c;
				riIndexChar++;
			}
		}
	}

	return 0;
}

void loop()
{
	char szLineString[READBUFFERSIZE];
	if( !ReadLineString( g_gps,
						 g_szReadBuffer, READBUFFERSIZE, g_iIndexChar,
						 szLineString, READBUFFERSIZE ) )
	{	// 読み取り途中
		return;
	}
	// 読み取り完了

	float fLat = 0.0;
	float fLong = 0.0;
	if( !AnalyzeLineString( szLineString, fLat, fLong ) )
	{
		return;
	}
	// 緯度、経度を読み取れた。

	// float to string
	char szLat[16];
	char szLong[16];
	dtostrf(fLat, 10, 6, szLat);
	dtostrf(fLong, 10, 6, szLong);

	// LCD出力
	char szBuffer[16];
	sprintf( szBuffer, "Lat. :%s", szLat );
	g_lcd.setCursor( 0, 0 );
	g_lcd.print( szBuffer );
	sprintf( szBuffer, "Long.:%s", szLong );
	g_lcd.setCursor( 0, 1 );
	g_lcd.print( szBuffer );
}

